import hashlib


def md5(path, block_size=64*1024):
    with open(path, 'rb') as f:
        md5 = hashlib.md5()
        while True:
            data = f.read(block_size)
            if not data:
                break
            md5.update(data)
    return md5.hexdigest()



import torch


def denormalize(x):
    mean = torch.tensor([[0.485, 0.456, 0.406]]).to(x.device).view(1, 3, 1, 1)
    std = torch.tensor([[0.229, 0.224, 0.225]]).to(x.device).view(1, 3, 1, 1)
    return x*std + mean



import torch.nn as nn


class IntersecLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        assert reduction in ('batchmean',)

    def forward(self, input, target):
        intersec = input.min(target).sum(dim=1).mean()
        return intersec



import torch.nn as nn


class CosineLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        assert reduction in('batchmean',)

    def forward(self, input, target):
        cosine = nn.functional.cosine_similarity(input, target, dim=1).mean()
        return cosine



import torch.nn as nn


class Accuracy(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        assert reduction in ('batchmean',)

    def forward(self, input, target):
        if len(target.shape) > 1:
            target = target.argmax(dim=1)
        _, input = input.max(dim=1)
        acc = input.eq(target).sum().float()/target.size(0)
        return acc



import os
import tarfile
import re
from .hash import md5


def _safeopen(path):
    """ Check if the path exists, if not, create it.
    """
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def compress(src, save_root, tar_name=None, ignore=None):
    """ Generating a compressed package using tar.bz2 (for md5 timestep issue).
        If src is a directory, the entire source inside the root would be compressed.
        The tar.bz2 file is named with md5 hash by default.
        param ignore: a regex pattern to ignore some files. (works only in directory mode)
    """
    tmp_path = _safeopen(save_root) + '/_tmp_compress.tar.bz2'
    with tarfile.open(tmp_path, 'w:bz2') as tar:
        if os.path.isdir(src):
            for dir_path, _, names in os.walk(src):
                for name in names:
                    full_path = os.path.join(dir_path, name)
                    if ignore is None or not re.match(re.compile(ignore), full_path):
                        print(f'add: {full_path}')
                        tar.add(full_path, os.path.join(dir_path.replace(src, ''), name))
                    else:
                        print(f'ignore: {full_path}')
        else:
            tar.add(src, os.path.split(src)[1])
    if tar_name is None:
        tar_name = str(md5(tmp_path))
    os.rename(tmp_path, tmp_path.replace(os.path.split(tmp_path)[1], tar_name + '.tar.bz2'))
    return tar_name



import base64
import requests
from bs4 import BeautifulSoup

url = 'http://@/include/auth_action.php'

headers = {"Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8",
           "Accept-Encoding": "gzip, deflate",
           "Accept-Language": "en-US,en;q=0.5",
           "Cache-Control": "max-age=0",
           "Connection": "keep-alive",
           "Host": "172.20.3.81:802",
           "Upgrade-Insecure-Requests": "1",
           "User-Agent": "Mozilla/5.0 (X11; Linux x86_64; rv:71.0) Gecko/20100101 Firefox/71.0"
           }


def post(url, d):
    return requests.post(url=url, data=d, headers=headers)


def logout(username, password):
    r = requests.get('http://login.ecust.edu.cn', allow_redirects=False)
    yield f':: Requesting login.ecust.edu.cn...\n response as:\t{r}'
    if r.status_code != 302:
        redirect_url = BeautifulSoup(r.content, 'lxml').find('meta').get('content').split('url=')[-1]
        r = requests.get(redirect_url)
        r = r.history[0]
        yield f'(1/2) forced redirect:\t{redirect_url}'
    else:
        yield f'(1/2) performing redirect'
    ip = r.headers['Location'].split('//')[-1].split('/')[0]
    yield f'(2/2) target ip:\t{ip}'
    data = {"action": "logout", "username": username, "password": password, "ajax": "1"}
    yield f":: Processing logout...\n:: Sending post...\n\nresponse:\t{post(url.replace('@', ip), data).content.decode('utf-8')}\n"


def login(username, password, free=True):
    if free:
        username = f'{username}@free'
    yield f':: Encrypting password...'
    bs = base64.b64encode(password.encode("utf8"))
    password = '{B}'+str(bs).replace('b\'', '').replace('\'', '')
    yield f'(1/1) generate base64-utf8 password'
    r = requests.get('http://login.ecust.edu.cn', allow_redirects=False)
    yield f':: Requesting login.ecust.edu.cn...\n response as:\t{r}'
    if r.status_code != 302:
        redirect_url = BeautifulSoup(r.content, 'lxml').find('meta').get('content').split('url=')[-1]
        r = requests.get(redirect_url)
        r = r.history[0]
        yield f'(1/3) forced redirect:\t{redirect_url}'
    else:
        yield f'(1/3) performing redirect'
    ip = r.headers['Location'].split('//')[-1].split('/')[0]
    ac_id = r.headers['Location'].split('/index_')[-1].split('.')[0]
    yield f'(2/3) target ip:\t{ip}\n(3/3) get ac_id as:\t{ac_id}'
    data = {"action": "login", "username": username, "password": password, "ac_id": ac_id, "user_ip": "", "nas_ip": "", "user_mac": "", "ajax": "1"}
    yield f"\nresponse:\t{post(url.replace('@', ip), data).content.decode('utf-8')}\n"


def test_connection():
    try:
        test_url = 'http://www.baidu.com'
        print(f':: Test connection...\n:: Request from {test_url}...')
        r = requests.get(test_url, timeout=2)
        if "<html>\n<head>\n<meta http-equiv='refresh' content='1; url=http://www.baidu.com/&arubalp=" in r.text:
            print(' response: redirected')
            return False
        print(f' response:\t{r}')
        return True
    except Exception as e:
        print(e)
        print(f' response:\tconnection fail')
        return False



import torch
from torch.utils.data import DataLoader


def save_checkpoint(model, optimizer, file_name='my_checkpoint.pth.tar'):
    print("=> Saving checkpoint")
    torch.save(
        {
            'model': model.state_dict(),
            'optimizer': optimizer.state_dict(),
        },
        file_name,
    )


def load_checkpoint(model, optimizer, file_name='my_checkpoint.pth.tar'):
    print("=> Loading checkpoint")
    checkpoint = torch.load(file_name)
    model.load_state_dict(checkpoint['model'])
    optimizer.load_state_dict(checkpoint['optimizer'])


def check_loss(dataloader, model, loss_fn, device):
    model.eval()
    loss = 0

    for data, targets in dataloader:
        data = data.to(device=device)
        targets = targets.to(device=device)

        with torch.no_grad():
            predictions = model(data)
            loss += loss_fn(predictions, targets).item()

    print(
        f'Average {loss_fn} over {len(dataloader)} batches: {loss/len(dataloader):.4f}'
    )
    model.train()
    return loss


def get_loaders(trainset, valset, batch_size, num_workers, pin_memory):

    trainloader = DataLoader(
        dataset=trainset,
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )

    valloader = DataLoader(
        dataset=valset,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )

    return trainloader, valloader



import numpy as np
import cv2
import torch


def heatmap_numpy(tensor, colormap=cv2.COLORMAP_JET):
    assert tensor.shape[0] == 1 and tensor.shape[1] == 1
    tensor = (tensor - tensor.min())/(tensor.max() - tensor.min())
    tensor = tensor.view(*tensor.shape[-2:])
    heatmap = cv2.applyColorMap(np.uint8(255 * tensor), colormap)
    return heatmap


def heatmap(tensor, colormap=cv2.COLORMAP_JET):
    ndarr = heatmap_numpy(tensor, colormap)
    tensor = torch.from_numpy(ndarr)
    tensor = tensor.permute(2, 0, 1).unsqueeze(0)
    return tensor



from torch.nn import *

from .accuracy import Accuracy
from .canber import CanberLoss
from .chebyshev import ChebyshevLoss
from .clark import ClarkLoss
from .correlation_coefficients import CorrelationCoefficientsLoss
from .cosine import CosineLoss
from .intersec import IntersecLoss
from .spatial_softmax2d import SpatialSoftmax2d
from .input_mask import InputMask

from .denormalize import denormalize
from .heatmap import heatmap, heatmap_numpy



import pynvml


def gpu_stat(index=0):
    pynvml.nvmlInit()
    handle = pynvml.nvmlDeviceGetHandleByIndex(index)  # 0表示显卡标号
    meminfo = pynvml.nvmlDeviceGetMemoryInfo(handle)
    return {
        'total': str(meminfo.total/1024**2)+'MiB',
        'used': str(meminfo.used/1024**2)+'MiB',
        'free': str(meminfo.free/1024**2)+'MiB',
    }




import torch.nn as nn


class InputMask(nn.Module):
    def __init__(self, module, module_pred=None, module_target=None):
        super().__init__()
        self.module_inner = module
        self.module_pred = module_pred
        self.module_target = module_target

    def forward(self, inputs, targets):
        arg1 = inputs if self.module_pred is None else self.module_pred(inputs)
        arg2 = targets if self.module_target is None else self.module_target(targets)
        return self.module_inner(arg1, arg2)



import torch
import torch.nn as nn


class CorrelationCoefficientsLoss(nn.Module):
    def __init__(self, reduction='batchmean'):
        super().__init__()
        assert reduction in ('batchmean',)

    def forward(self, input, target):
        batch_size = input.shape[0]
        input = input.view(input.size(0), -1)
        target = target.view(target.size(0), -1)
        CC = []
        for i in range(batch_size):
            im = input[i] - torch.mean(input[i])
            tm = target[i] - torch.mean(target[i])
            CC.append(torch.sum(im * tm) / (torch.sqrt(torch.sum(im ** 2))
                                            * torch.sqrt(torch.sum(tm ** 2))))
            CC[i].unsqueeze_(0)
        CC = torch.cat(CC, 0)
        CC = torch.mean(CC)
        return CC
